#!/usr/bin/env python

docstring='''spusteni 
wanna.py [datovy soubor] -par=hodnota  -par2=hodnota2 ...

kde par muze byt 
a : polovina vzdalenosti elektrod
R : polomer elektrod
U : napeti mezi nimi
V : seznam napeti ekvipot. hladin oddelenych carkami (bez mezer)

funkce promptu:
plot: vykresli znovu
cent: vypocet stredu kruznic
fit: optimalni hodnoty parametru a,R,U pro shodu teorie s merenim
recon: jaka napeti odpovidaji ekvipot. hladinam pro dane a,R,U?
param: vypis parametru

R=hodnota apod : nastaveni parametru

'''

import numpy as np
from matplotlib import pyplot as plt

par={}
par['a']=10
par['U']=5
par['R']=2.5

legend = False #bezna legenda matplotlibu nebo popisky primo u car
colors = ['red', 'orange', 'yellow', 'green', 'cyan', 'blue', 'purple', 'pink', 'brown','navy']

def diff_me(hladina,x,y,U=5,a=17.5,R=2.5,V=[],glev=None,doplot=True):
    '''vykresli kruznice a merene body 
    nebo jen dopocita vzdalenosti mezi nimi'''
    #vytvorenie masky pre kazdu hladinu
    h = np.sqrt(a**2+R**2)
    
    nlev=int(hladina.max() + 1)
    if not np.iterable(glev): glev=range(nlev)
    hladina_masky = []
    for i in glev:
        hladina_masky.append(hladina == i)

    if len(V)==0:
        V=np.linspace(0,abs(U),nlev+2)[1:-1]
    #V = np.array([1.5, 2, 2.5, 3, 3.5, 0.5, 1, 4, 4.5]) #upravit podla vlastneho merania

    #vypocet velicin pomocou numpy
    lam = np.exp(((V/abs(U))-0.5)*2*(np.log((h+a)/R)))
    ys = a*(((lam**2)+1)/((lam**2)-1))
    r = np.sqrt(ys**2-a**2)

    #vykreslenie teoretickych kruznic
    theta = np.linspace(0, 2 * np.pi, 100)

    if doplot: fig, ax = plt.subplots()
    else: res=0
    pos=[]
    ymin,ymax=np.min(y)-2,np.max(y)+2
    xmin,xmax=np.min(x)-2, np.max(x)+2
    
    for i in range (len(V)):
        if i>=len(hladina_masky) or sum(hladina_masky[i])==0: continue
        if sum(hladina_masky[i])<2: continue
        if not doplot: #jen pocita residua
            if r[i]==np.inf: continue
            dis=np.sqrt(x[hladina_masky[i]]**2+ (y[hladina_masky[i]]-ys[i])**2)
            res+=np.sum((dis-r[i])**2)
        else: #vykresleni 
            ax.plot(x[hladina_masky[i]], y[hladina_masky[i]], '.', color=colors[i])
            if r[i]==np.inf or V[i]==U/2: #rovna cara uprostred
                plt.axhline(0,xmin,xmax,color=colors[i])
                if not legend:
                    plt.text(xmax+0.5,1,'%.2f V'%V[i])
                continue
            circle_x = r[i] * np.cos(theta)
            circle_y = ys[i] + (r[i] * np.sin(theta))
            ax.plot(circle_x, circle_y, color=colors[i],label='%.2f V'%V[i])
            pos.append([circle_x[::10],circle_y[::10]])
            if legend==False: #popisky primo u car
                sel=(circle_x>xmin)*(circle_x<xmax)
                sel*=(circle_y>ymin)*(circle_y<ymax)
                if ys[i]>0:
                    yq=circle_y[sel].max()-2
                    xq=circle_x[sel][circle_y[sel]>yq]
                    xq=xq.min()
                    if yq>ymax: yq=ymax
                else:
                    yq=circle_y[sel].min()+2
                    xq=circle_x[sel][circle_y[sel]<yq]
                    xq=xq.min()
                    if yq<ymin: yq=ymin+0.2
                plt.text(xq,yq,'%.2f V'%V[i])
                #print(xq,yq,'%.2f V'%V[i])
    
    if not doplot: return res
    ax.set_aspect('equal')
    fig.set_size_inches(6, 6)
    plt.xlim([xmin,xmax])
    plt.ylim([ymin,ymax])
    ax.set_xlabel('x [cm]')
    ax.set_ylabel('y [cm]')
    if legend: ax.legend()

    plt.show()

#----------------------------------------------------

def triang(x1,x2,x3):
    '''stred kruznice opsane trojuhelniku'''
    m=(x1+x2)/2
    p=np.r_[(x2[1]-x1[1])/2,(x1[0]-x2[0])/2]
    m2=(x3+x2)/2
    p2=np.r_[(x2[1]-x3[1])/2,(x3[0]-x2[0])/2]
    cxx=lambda a,b:a[1]*b[0]-a[0]*b[1]
    tm=(-cxx(m,p2)+cxx(m2,p2))/cxx(p,p2)
    return m+tm*p

def get_cent(x,y,refine=False,reject=2,full_output=False):
    '''nafitovani kruznice na sadu bodu
    vychozi odhad se ziska z mnoziny stredu mnoha moznych trojuhelniku
    '''
    import numpy as np
    zlst=[]
    comb=[]
    for i in range(len(x)-2):
        pts=[np.r_[x[i],y[i]]]
        for j in range(i+1,len(x)-1):
            pts=pts[:1]+[np.r_[x[j],y[j]]]
            for k in range(j+1,len(x)):
                pts=pts[:2]+[np.r_[x[k],y[k]]]
                zlst.append(triang(*pts))
                comb.append([i,j,k])
    if not refine:
        return zlst

    from scipy.cluster.vq import whiten
    from scipy.spatial import distance
    whit2=whiten(zlst)
    cini=np.median(whit2,0)
    from scipy.optimize import minimize
    res=minimize(lambda c:distance.cdist([c],whit2).sum(), cini, method='L-BFGS-B')
    cent=res.x
    dist=distance.cdist([cent],whit2)[0]
    allids=np.array(comb)[dist<2*np.median(dist)].ravel()

    from collections import Counter
    cts=Counter(allids)
    bestids=np.argsort([cts[i] for i in range(len(cts))])[::-1]
    #return bestids
    new_cent=get_cent(x[bestids[:-reject]],y[bestids[:-reject]])
    if full_output: return new_cent,bestids[:-reject]
    return new_cent

def all_cents(hladina,x,y,refine=True):
    '''najde stredy vsech ekvipotenc. kruznic v datovem souboru
    '''
    from scipy.spatial import distance
    rep=[]
    nlev=int(hladina.max() + 1)
    for i in range(nlev):
        sq=hladina==i
        if sum(sq)<3:
            rep.append((0,[]))
        else:
            if refine:
                from scipy.optimize import minimize
                cals=get_cent(x[sq],y[sq],refine=True,full_output=True)
                cent=np.median(cals[0],0)
                sprd=lambda cent:distance.cdist([cent],np.array([x[sq],y[sq]]).T[cals[1]])[0].std()
                new_cent=minimize(sprd,cent).x
                rep.append((len(cals),new_cent,distance.cdist([new_cent],np.array([x[sq],y[sq]]).T[cals[1]])[0].mean()))
            else:
                cals=get_cent(x[sq],y[sq],refine=False)
                rep.append((len(cals),np.median(cals,0)))
    return rep
#-----------------------------------------------------------

l2volts=lambda lame,a,R,U:np.log(lame)/2/np.log((a+np.sqrt(a**2+R**2))/R)*U+U/2
def recon_volts(hl,xl,yl):
    '''vypocet napeti z parametru kruznic'''
    cts=all_cents(hl,xl,yl)
    ye=np.array([c[1][1] for c in cts])
    re=np.array([c[2] for c in cts])
    ae=np.sqrt(np.median(ye**2-re**2))
    #assert np.all(abs(ye)>ae)
    return l2volts(np.sqrt((ye/ae+1)/(ye/ae-1)),ae,par['R'],par['U'])


def process(hl,xl,yl,V=[]):
    '''optimalizace parametru 
    '''
    cts=all_cents(hl,xl,yl)
    nwo=np.argsort([np.sign(c[1][1])/c[2] if len(c)>2 else 100 for c in cts])
    print("order:",nwo)
    funx=lambda p:diff_me(hl,xl+p[3],yl+p[4],V=V,glev=list(nwo),a=p[0],R=p[1],U=p[2],doplot=False)
    from scipy import optimize as op
    out=op.fmin(funx,[par['a'],par['R'],par['U'],0,0])

    diff_me(hl,xl+out[3],yl+out[4],a=out[0],R=out[1],U=out[2],V=V,glev=list(nwo))
    labs=["distance","R_probe","Voltage","X corr","Y corr"]
    return list(zip(labs,out))

#-----------------------------------------------------------

import sys,os

def par_se(v):
    a,b=v.split('=')
    if a=='a': par['a']=float(b)
    elif a=='U': par['U']=float(b)
    elif a=='R': par['R']=float(b)
    elif a=='V': 
        par['V']=np.array([float(c) for c in b.strip().split(',')])

fname=''
if __name__ == '__main__':
    for v in sys.argv[1:]:
        if v[0]=='-': par_se(v[1:])
        elif os.path.exists(v):
            fname=v

    if fname=='':
        print(docstring)
        while not os.path.exists(fname):
            fname=input('Jmeno datoveho souboru:')
    hladina,x,y=np.loadtxt(fname).T
    if 'V' in par:
        par['V']=par['V'][par['V']<par['U']]
        print(len(par['V']),' levels')
    print('obrazek lze ulozit ikonou diskety')
    print('pokracujeme po zavreni okna')
    diff_me(hladina,x,y,doplot=True,**par)#a=par_a, R=par['R'], U=par_U,
    while True:
        step=input("...jak dal? [fit, end, plot, param, cent, recon, {par}=]")
        if step=='fit':
            res=process(hladina,x,y,V=par.get('V',[]))
            print(res)
        elif step=='end':
            break
        elif step=='cent':
            cts=all_cents(hladina,x,y)
            print("Stredy a polomery kruznic")
            for i,c in enumerate(cts):
                if len(c)>2:
                    print(i,': [%8.4f %8.4f] '%tuple(c[1]),c[2])
        elif step[:5]=='param':
            for a,b in par.items():
                print(a,': ',b)
        elif step=='recon':
            trueV=recon_volts(hladina,x,y)
            print("Idealni napeti odpovidajici hladinam:")
            for i in range(len(trueV)):
                print(f"{i}: {trueV[i]:.4f}V")
                if step.find('save')>0: par['V']=trueV
        elif step.find('=')>0:
            par_se(step)
            if 'V'==step[0]:
                par['V']=par['V'][par['V']<par['U']]
                print(len(par['V']),' levels')
        elif step=='plot':
            diff_me(hladina,x,y,doplot=True,**par)#a=par_a, R=par['R'], U=par_U,

    fname=''#input('ulozit PNG do souboru [prazny=bez ulozeni]?')
    if len(fname)>2:
        if fname[-4:] not in ['png','eps','jpg']: fname+='.png'
        plt.savefig(fname)
        
